/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "renderserver.h"
#include "gravitationallens.h"
#include <serut/memoryserializer.h>
#include <enut/packet.h>
#include <enut/tcpv4socket.h>
#include <enut/tcppacketsocket.h>
#include <signal.h>
#include <time.h>
#include <stdlib.h>
#include <iostream>
#include <string>

#include "debugnew.h"

using namespace grale;
bool endserverloop = false;

void SignalHandler(int val)
{
	endserverloop = true;
}

int main2(int argc, char *argv[])
{
	if (argc != 3)
	{
		std::cerr << "Usage: " << std::endl;
		std::cerr << std::string(argv[0]) << " serverIP portnumber" << std::endl; 
		return -1;
	}

	uint16_t port = (uint16_t)atoi(argv[2]);
	nut::IPv4Address address;
	nut::TCPv4Socket tcpSocket;
	nut::TCPPacketSocket connection(&tcpSocket, false, false, RENDERSERVER_MAXPACKSIZE, sizeof(uint32_t), RENDERSERVER_PACKID);

	address.setAddress(argv[1]);
	
	if (!tcpSocket.create())
	{
		std::cerr << "Error: " << std::endl;
		std::cerr << " Couldn't create TCP socket: " << tcpSocket.getErrorString() << std::endl;
		return -1;
	}

	std::cout << "Connecting to server" << std::endl;
	if (!tcpSocket.connect(address, port))
	{
		std::cerr << "Error: " << std::endl;
		std::cerr << " Couldn't open connection: " << connection.getErrorString() << std::endl;
		return -1;
	}

	// Ok, connection opened, send identification

	{
		uint8_t buf[sizeof(int32_t)];
		serut::MemorySerializer mser(0, 0, buf, sizeof(int32_t));

		mser.writeInt32(RENDERSERVER_COMMAND_RENDERER);

		std::cout << "Sending identification" << std::endl;
		if (!connection.write(buf, sizeof(int32_t)))
		{
			std::cerr << "Error: " << std::endl;
			std::cerr << " Error writing identification packet: " << connection.getErrorString() << std::endl;
			return -1;
		}

	}

#ifndef WIN32
	std::cout << "Installing signal handlers" << std::endl;
	signal(SIGTERM,SignalHandler);
	signal(SIGHUP,SignalHandler);
	signal(SIGQUIT,SignalHandler);
	signal(SIGABRT,SignalHandler);
	signal(SIGINT,SignalHandler);
#endif // WIN32

	GravitationalLens *lens = 0;
	Vector2D<double> bottomleft;
	Vector2D<double> topright;
	double D_s = 0;
	double D_ds = 0;
	double dx = 0;
	double dy = 0;
	int numx = 0;
	int numy = 0;
	uint8_t *linebuffer = 0;
	int linebuffersize = 0;
	
	// Ok, wait for incoming data from the server
	while (!endserverloop)
	{
		if (!connection.waitForData(1, 0))
		{
			endserverloop = true;
			continue;
		}

		if (connection.isDataAvailable())
		{
			size_t length = 0;

			connection.getAvailableDataLength(length);

			if (length == 0)
			{
				std::cerr << "Error: " << std::endl;
				std::cerr << " Server closed connection. " << std::endl;
				return -1;
			}
			else
			{	
				if (!connection.poll())
				{
					std::cerr << "Error: " << std::endl;
					std::cerr << " Error polling for new data: " << connection.getErrorString() << std::endl;
					return -1;
				}
				
				nut::Packet packet;

				while (connection.read(packet))
				{
					if (packet.getLength() < 4)
					{
						std::cerr << "Error: " << std::endl;
						std::cerr << " Received invalid packet from server" << std::endl;
						return -1;
					}
					
					serut::MemorySerializer mser(packet.getData(), packet.getLength(), 0, 0);
					int32_t cmd;
					
					mser.readInt32(&cmd);

					if (cmd == RENDERSERVER_COMMAND_ECHO)
					{
						// just ignore it
					}
					else if (cmd == RENDERSERVER_COMMAND_RENDER)
					{
						//std::cout << "Received render request" << std::endl;
				
						if (packet.getLength() != sizeof(int32_t)*2)
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Received invalid render packet from server" << std::endl;
							return -1;
						}
						
						if (lens == 0)
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Received render packet from server but no lens information has been received yet" << std::endl;
							return -1;
						}
						
						int32_t line;

						mser.readInt32(&line);
						
						if (line < 0 || line >= numy)
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Received line number is out of range" << std::endl;
							return -1;
						}

						double x = bottomleft.getX();
						double y = bottomleft.getY()+(((double)line)*dy);
					
						serut::MemorySerializer mser2(0,0,linebuffer,linebuffersize);
						mser2.writeInt32(RENDERSERVER_COMMAND_RESULT);

						std::cout << "Rendering requested line " << line << std::endl;

						for (int i = 0 ; i < numx ; i++, x+= dx)
						{
							Vector2D<double> theta(x,y);
							Vector2D<double> alpha;
							double data[5];

							if (!lens->getAlphaVector(theta,&alpha))
							{
								std::cerr << "Error: " << std::endl;
								std::cerr << " Couldn't get deflection angle: " << lens->getErrorString() << std::endl;
								return -1;
							}
							
							alpha *= (D_ds/D_s);

							data[0] = alpha.getX();
							data[1] = alpha.getY();
							
							double axx,ayy,axy;
							lens->getAlphaVectorDerivatives(theta,axx,ayy,axy);
							data[2] = axx;
							data[3] = ayy;
							data[4] = axy;

							mser2.writeDoubles(data,5);

							if ((connection.getLastWriteTime() - time(0)) > 10)
							{
								uint8_t buf[sizeof(int32_t)];
								serut::MemorySerializer mser(0,0,buf,sizeof(int32_t));

								mser.writeInt32(RENDERSERVER_COMMAND_ECHO);

								//std::cout << "Sending echo to server" << std::endl;
								if (!connection.write(buf, sizeof(int32_t)))
								{
									std::cerr << "Error: " << std::endl;
									std::cerr << " Error writing echo packet: " << connection.getErrorString() << std::endl; 
									return -1;
								}
							}
						}
						
						// Calculation ended, write result

						//std::cout << "Sending rendered line to server" << std::endl;
						if (!connection.write(linebuffer, linebuffersize))
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Error writing render result: " << connection.getErrorString() << std::endl; 
							return -1;
						}
					}
					else if (cmd == RENDERSERVER_COMMAND_LENS)
					{
						//std::cout << "Received lens information" << std::endl;

						bool error = false;
						if (!mser.readInt32(&numx))
							error = true;
						if (!mser.readInt32(&numy))
							error = true;
						if (numx < 2 || numx > 100000)
							error = true;
						if (numy < 2 || numy > 100000)
							error = true;
						if (!mser.readDoubles(bottomleft.getComponents(), 2))
							error = true;
						if (!mser.readDoubles(topright.getComponents(), 2))
							error = true;
						if (!mser.readDouble(&D_s))
							error = true;
						if (!mser.readDouble(&D_ds))
							error = true;

						GravitationalLens *l = 0;
						std::string errstr;
				
						if (!GravitationalLens::read(mser,&l,errstr))
							error = true;
				
						if (mser.getBytesRead() != packet.getLength())
						{
							delete l;
							error = true;
						}
						
						if (error)
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Received invalid lens information from server" << std::endl;
							return -1;
						}

						// No error, store new lens

						if (lens)
						{
							delete lens;
							delete [] linebuffer;
						}
						
						lens = l;
						linebuffersize = sizeof(int32_t)+sizeof(double)*5*numx;
						linebuffer = new uint8_t[linebuffersize];

						dx = (topright.getX()-bottomleft.getX())/((double)(numx-1));
						dy = (topright.getY()-bottomleft.getY())/((double)(numy-1));

						// Inform server that we've received the lens

						uint8_t buf[sizeof(int32_t)];
						serut::MemorySerializer mser2(0,0,buf,sizeof(int32_t));

						mser2.writeInt32(RENDERSERVER_COMMAND_RECEIVEDLENS);

						if (!connection.write(buf, sizeof(int32_t)))
						{
							std::cerr << "Error: " << std::endl;
							std::cerr << " Error writing received lens acknowledgment: " << connection.getErrorString() << std::endl; 
							return -1;
						}
					}
					else
					{
						std::cerr << "Error: " << std::endl;
						std::cerr << " Received invalid command from server" << std::endl;
						return -1;
					}
				}
			}
		}

		time_t curtime = time(0);	

		if ((curtime - connection.getLastWriteTime()) > 10) // write something every 10 seconds
		{
			uint8_t buf[sizeof(int32_t)];
			serut::MemorySerializer mser(0,0,buf,sizeof(int32_t));

			mser.writeInt32(RENDERSERVER_COMMAND_ECHO);

			if (!connection.write(buf, sizeof(int32_t)))
			{
				std::cerr << "Error: " << std::endl;
				std::cerr << " Error writing echo packet: " << connection.getErrorString() << std::endl; 
				return -1;
			}
		}

		if ((curtime - connection.getLastReadTime()) > 1200) // server died
		{
			std::cerr << "Error: " << std::endl;
			std::cerr << " Server didn't send anything for the last minute" << std::endl; 
			return -1;
		}
	}

	if (linebuffer)
		delete [] linebuffer;
	if (lens)
		delete lens;
	
	return 0;
}

int main(int argc, char *argv[])
{
#ifdef WIN32
	WSADATA dat;
	WSAStartup(MAKEWORD(2,2),&dat);
#endif // WIN32
	int status = main2(argc, argv);
#ifdef WIN32
	WSACleanup();
#endif
	return status;
}