/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "cosmology.h"
#include "realtype.h"
#include "constants.h"

#include "debugnew.h"

namespace grale
{

Cosmology::Cosmology(double h, double Wm, double Wr, double Wv, double w)
{
	m_h = h;
	m_Wm = Wm;
	m_Wr = Wr;
	m_Wv = Wv;
	m_w = w;
}

Cosmology::~Cosmology()
{
}

double Cosmology::getAngularDiameterDistance(double z1, double z2, int integrationSteps) const
{
	if (integrationSteps < 1)
		integrationSteps = 1;

	if (z1 > z2)
	{
		double temp = z1;
		z1 = z2;
		z2 = temp;
	}

	double Wk = 1.0 - m_Wv - m_Wr - m_Wm;
	double sum = 0;
	double corrsum = 0;
	double R2 = 1.0/(1.0+z1);
	double R1 = 1.0/(1.0+z2);
	double dR = (R2-R1)/((double)integrationSteps);
	double R = R1;
	
	for (int i = 0 ; i < integrationSteps ; i++, R += dR)
	{
		double r1 = R;
		double r2 = R+dR;
		
		double powr = POW(r1,1.0-3.0*m_w);
		double term1 = m_Wv*powr;
		double term2 = m_Wm*r1;
		double term3 = m_Wr;
		double term4 = Wk*r1*r1;
		double termsum = term1+term2+term3+term4;

		double val1 = 1.0/SQRT(termsum);
		double deriv1 = -0.5*(m_Wm*r1+m_Wv*powr-3.0*m_Wv*powr*m_w+2.0*Wk*r1*r1)/(POW(termsum,1.5)*r1);
		
		powr = POW(r2,1.0-3.0*m_w);
		term1 = m_Wv*powr;
		term2 = m_Wm*r2;
		term3 = m_Wr;
		term4 = Wk*r2*r2;
		termsum = term1+term2+term3+term4;

		double val2 = 1.0/SQRT(termsum);
		double deriv2 = -0.5*(m_Wm*r2+m_Wv*powr-3.0*m_Wv*powr*m_w+2.0*Wk*r2*r2)/(POW(termsum,1.5)*r2);
		
		double scale = ABS(val1);
		double q1 = (val2-val1)/scale;
		double dq0 = deriv1/scale*dR;
		double dq1 = deriv2/scale*dR;

		double a = -2.0*q1+dq0+dq1;
		double b = 3.0*q1-2.0*dq0-dq1;
		double c = dq0;

		corrsum += (a/4.0+b/3.0+c/2.0)*scale*dR;
		sum += val1*dR;
	}

	sum += corrsum;

	double A = 0;
	
	if (Wk == 0)
		A = sum;
	else
	{
		double prod = SQRT(ABS(Wk))*sum;

		if (ABS(prod) < 1e-5) // flat enough
			A = sum;
		else
		{
			if (Wk > 0)
				A = (1.0/SQRT(Wk))*SINH(prod);
			else // Wk < 0
				A = (1.0/SQRT(-Wk))*SIN(prod);
		}
	}
	
	return (SPEED_C*A*((1.0/(1.0+z2))/100000.0)/m_h)*DIST_MPC;
}

} // namespace

