/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "arealensinversiongafactoryparams.h"
#include "imagesdata.h"

#include "debugnew.h"

namespace grale
{

AreaLensInversionGAFactoryParams::AreaLensInversionGAFactoryParams() : LensInversionGAFactoryParams(LensInversionGAFactoryParams::AreaInversion)
{
	deleteimages = false;
	D_d = 0;
	maxgenerations = 0;
	areasize = 0;
}

AreaLensInversionGAFactoryParams::~AreaLensInversionGAFactoryParams()
{
	if (deleteimages)
	{
		while (!images.empty())
		{
			delete (*(images.begin()));
			images.pop_front();
		}
	}
}

void AreaLensInversionGAFactoryParams::init(int maxgen, const std::list<ImagesDataExtended *> &images, 
								   double D_d, Vector2D<double> center, double areasize, bool copyimages)
{
	if (deleteimages)
	{
		while (!AreaLensInversionGAFactoryParams::images.empty())
		{
			delete (*(AreaLensInversionGAFactoryParams::images.begin()));
			AreaLensInversionGAFactoryParams::images.pop_front();
		}
	}
	AreaLensInversionGAFactoryParams::images.clear();	
	
	if (copyimages)
		deleteimages = true;
	else
		deleteimages = false;

	maxgenerations = maxgen;

	AreaLensInversionGAFactoryParams::D_d = D_d;
	AreaLensInversionGAFactoryParams::center = center;
	AreaLensInversionGAFactoryParams::areasize = areasize;
	
	std::list<ImagesDataExtended *>::const_iterator it;

	for (it = images.begin() ; it != images.end() ; it++)
	{
		ImagesDataExtended *img = (*it);
		ImagesDataExtended *img2;

		if (copyimages)
			img2 = new ImagesDataExtended(*img);
		else
			img2 = img;
			
		AreaLensInversionGAFactoryParams::images.push_back(img2);
	}
}

bool AreaLensInversionGAFactoryParams::write(serut::SerializationInterface &si) const
{
	if (!si.writeInt32(maxgenerations))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(D_d))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDouble(areasize))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.writeDoubles(center.getComponents(), 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t sizes[1];

	sizes[0] = images.size();

	if (!si.writeInt32s(sizes,1))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	std::list<ImagesDataExtended *>::const_iterator it;

	for (it = images.begin() ; it != images.end() ; it++)
	{
		double distances[2];

		distances[0] = (*it)->getD_s();
		distances[1] = (*it)->getD_ds();

		if (!si.writeDoubles(distances,2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		if (!(*it)->write(si))
		{
			setErrorString((*it)->getErrorString());
			return false;
		}
	}

	if (!writeInternal(si))
		return false;
	if (!si.writeInt32(-1)) // write marker
	{
		setErrorString(si.getErrorString());
		return false;
	}
	return true;
}

bool AreaLensInversionGAFactoryParams::read(serut::SerializationInterface &si)
{
	if (deleteimages)
	{
		while (!images.empty())
		{
			delete (*(images.begin()));
			images.pop_front();
		}
	}
	images.clear();
	deleteimages = true;
	
	int32_t maxgen;

	if (!si.readInt32(&maxgen))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	maxgenerations = (int)maxgen;
	
	if (!si.readDouble(&D_d))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDouble(&areasize))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readDoubles(center.getComponents(), 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t sizes[1];

	if (!si.readInt32s(sizes,1))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t numimages = sizes[0];
	
	for (int32_t i = 0 ; i <numimages ; i++)
	{
		double distances[2];

		if (!si.readDoubles(distances,2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		ImagesDataExtended *img = new ImagesDataExtended(distances[0],distances[1]);
		if (!img->read(si))
		{
			setErrorString(img->getErrorString());
			delete img;
			return false;
		}
		images.push_back(img);
	}

	if (!readInternal(si))
		return false;

	int32_t marker = 0;
	if (!si.readInt32(&marker))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (marker != -1)
	{
		setErrorString("Read invalid marker value");
		return false;
	}
	return true;
}

bool AreaLensInversionGAFactoryParams::loadExtraParameters(const std::string &fname)
{
	setErrorString("This implementation does not need extra parameters");
	return false;
}
	
} // end namespace

