/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "gridlensinversiongenomebase.h"
#include "gridlensinversiongafactorybase.h"
#include "gridlensinversiongafactoryparams.h"
#include "multipleplummerlens.h"
#include "multiplesquarelens.h"
#include "multiplegausslens.h"
#include "doublepeakdistribution.h"
#include "uniformdistribution.h"
#include "simpleuniformdistribution.h"
#include "gaussiandistribution.h"
#include "powerlawdistribution.h"
#include "constants.h"
#include "lensfitnessobject.h"
#include "deflectionmatrix.h"
#include "backprojectmatrixnew.h"
#include <iostream>

#include "debugnew.h"

#define AUTODISTFRAC_MINVALUE 0.3f
// #define EXPCROSS

namespace grale
{

GridLensInversionGenomeBase::GridLensInversionGenomeBase(GridLensInversionGAFactoryBase *f, int nummasses, bool useGenomeMassSheet)
{
	factory = f;
	m_fitnessComp = 0;
	m_sheetFactor = 0;
	
	for (int i = 0 ; i < GRIDLENSINVERSIONGENOMEBASE_MAXFITNESSCOMP ; i++)
		fitnessvalues[i] = 0;
	
	scalefactor = 1.0;
	m_masses.resize(nummasses);

	SimpleUniformDistribution uniDist(factory->getRandomNumberGenerator());

	for (int i = 0 ; i < nummasses ; i++)
		m_masses[i] = (float)uniDist.pickNumber();

	if (factory->allowNegativeValues())
	{
		for (int i = 0 ; i < nummasses ; i++)
			m_masses[i] -= 0.5;
	}

	if (useGenomeMassSheet)
		m_sheetValue = uniDist.pickNumber();
	else
		m_sheetValue = -1;

#ifdef AUTODISTFRAC
	m_distanceFractions.resize(factory->getNumberOfAutoDistanceFractions());

	for (int i = 0 ; i < m_distanceFractions.size() ; i++)
		m_distanceFractions[i] = uniDist.pickNumber();
#endif // AUTODISTFRAC
	rescale();
}

#ifdef AUTODISTFRAC
GridLensInversionGenomeBase::GridLensInversionGenomeBase(GridLensInversionGAFactoryBase *f, const std::vector<float> &masses, const std::vector<float> &fractions, float sheetValue)
#else
GridLensInversionGenomeBase::GridLensInversionGenomeBase(GridLensInversionGAFactoryBase *f, const std::vector<float> &masses, float sheetValue)
#endif // AUTODISTFRAC
{
	factory = f;
	m_masses = masses;
	m_sheetValue = sheetValue;
	m_sheetFactor = 0;
#ifdef AUTODISTFRAC
	m_distanceFractions = fractions;
#endif // AUTODISTFRAC
	
	m_fitnessComp = 0;
	
	for (int i = 0 ; i < GRIDLENSINVERSIONGENOMEBASE_MAXFITNESSCOMP ; i++)
		fitnessvalues[i] = 0;

	scalefactor = 1.0;

	rescale();
}

GridLensInversionGenomeBase::~GridLensInversionGenomeBase()
{
}
	
bool GridLensInversionGenomeBase::calculateFitness()
{
	int nummasses = m_masses.size();
	float sheetFactor = 0;
	float startValue, startValue0;
	float stopValue, stopValue0;
	float sheetStartValue, sheetStartValue0;
	float sheetStopValue, sheetStopValue0;

	int numiterations;
	int numiterationsteps;

	initializeNewCalculation();

	// TODO: adjust mass scale depending on sheet value?

	if (factory->allowNegativeValues())
	{
		// TODO: is this enough? or do we need some additional constraints?

		float massum = 0;
		const float *weights = factory->getMassWeights();
		
		// TODO: should find something better!!
		for (int i = 0 ; i < nummasses ; i++)
			massum += ABS(m_masses[i]*weights[i]);

		getCalculationParameters(startValue, stopValue, numiterationsteps, numiterations);
		getSheetCalculationParameters(sheetStartValue, sheetStopValue);

		// we want a scale factor of 1 to correspond to the total mass specified by the
		// massscale used in the BackProjectMatrix

		startValue /= massum;
		stopValue /= massum;
	}
	else
	{
		float massum = 0;
		const float *weights = factory->getMassWeights();

		for (int i = 0 ; i < nummasses ; i++)
			massum += m_masses[i]*weights[i];

		getCalculationParameters(startValue, stopValue, numiterationsteps, numiterations);
		getSheetCalculationParameters(sheetStartValue, sheetStopValue);

		// we want a scale factor of 1 to correspond to the total mass specified by the
		// massscale used in the BackProjectMatrix

		startValue /= massum;
		stopValue /= massum;
	}

	bool loopSheet = factory->useLoopSheet();

	startValue0 = startValue;
	stopValue0 = stopValue;

	if (!loopSheet)
	{
		float currentBestFitness = 1e20f;
		float currentBestScaleFactor = 1.0f;

		sheetFactor = factory->getSheetScale()*m_sheetValue;		

 		for (int i = 0 ; i < numiterations ; i++)
		{
			float stepsize = (stopValue-startValue)/((float)(numiterationsteps-1));
			float s = startValue;

			for (int j = 0 ; j < numiterationsteps ; j++, s += stepsize)
			{
				float f = calculateMassScaleFitness(s, sheetFactor);

				if (f < currentBestFitness)
				{
					currentBestFitness = f;
					currentBestScaleFactor = s;
				}
			}

			startValue = currentBestScaleFactor-stepsize;
			stopValue = currentBestScaleFactor+stepsize;

			if (startValue < startValue0)
				startValue = startValue0;
			if (stopValue > stopValue0)
				stopValue = stopValue0;
		}

		scalefactor = currentBestScaleFactor;
	}
	else
	{
		//std::cerr << std::endl << std::endl << std::endl;

		float currentBestFitness = 1e20f;
		float currentBestScaleFactor = 1.0f;
		float currentBestSheetFactor = 1.0f;
		float sheetScale = factory->getSheetScale();

		sheetStartValue *= sheetScale;
		sheetStopValue *= sheetScale;

		sheetStartValue0 = sheetStartValue;
		sheetStopValue0 = sheetStopValue;

 		for (int i = 0 ; i < numiterations ; i++)
		{
			float stepsize = (stopValue-startValue)/((float)(numiterationsteps-1));
			float sheetStepSize = (sheetStopValue-sheetStartValue)/((float)(numiterationsteps-1));
			float s = startValue;

			for (int j = 0 ; j < numiterationsteps ; j++, s += stepsize)
			{
				float s2 = sheetStartValue;

				for (int k = 0 ; k < numiterationsteps ; k++, s2 += sheetStepSize)
				{
					float f = calculateMassScaleFitness(s, s2);
				
					//std::cerr << s << " " << s2 << " " << f << std::endl;

					if (f < currentBestFitness)
					{
						currentBestFitness = f;
						currentBestScaleFactor = s;
						currentBestSheetFactor = s2;
					}
				}
				//std::cerr << std::endl;
			}

			startValue = currentBestScaleFactor-stepsize;
			stopValue = currentBestScaleFactor+stepsize;
			sheetStartValue = currentBestSheetFactor-sheetStepSize;
			sheetStopValue = currentBestSheetFactor+sheetStepSize;

			if (startValue < startValue0)
				startValue = startValue0;
			if (stopValue > stopValue0)
				stopValue = stopValue0;
			if (sheetStartValue < sheetStartValue0)
				sheetStartValue = sheetStartValue0;
			if (sheetStopValue > sheetStopValue0)
				sheetStopValue = sheetStopValue0;
		}

		scalefactor = currentBestScaleFactor;
		sheetFactor = currentBestSheetFactor;

		// std::cout << "Best factors are " << scalefactor << "," << sheetFactor << std::endl;

		m_sheetFactor = sheetFactor;
	}

	// Do the final (possibly multi-component) fitness evaluation
	calculateTotalFitness(scalefactor, sheetFactor, fitnessvalues);

	return true;
}

mogal::Genome *GridLensInversionGenomeBase::reproduce(const mogal::Genome *g) const
{
	int nummasses = m_masses.size();
	std::vector<float> newmasses(nummasses);
	const GridLensInversionGenomeBase *g2 = (const GridLensInversionGenomeBase *)g;
	SimpleUniformDistribution uniDist(factory->getRandomNumberGenerator());

	if (factory->allowNegativeValues())
	{
#ifndef EXPCROSS
		for (int i = 0 ; i < nummasses ; i++)
		{
			float x = (float)uniDist.pickNumber();

			if (x < 0.5f)
				newmasses[i] = scalefactor*m_masses[i];
			else
				newmasses[i] = (g2->scalefactor)*(g2->m_masses[i]);
		}
#else
		UniformDistribution massPick(0, nummasses, factory->getRandomNumberGenerator());

		int m1 = (int)massPick.pickNumber();
		int m2 = (int)massPick.pickNumber();

		if (m1 > m2)
		{
			int tmp = m2;
			m2 = m1;
			m1 = tmp;
		}

		for (int i = 0 ; i < m1 ; i++)
			newmasses[i] = scalefactor*m_masses[i];
		for (int i = m1 ; i < m2 ; i++)
			newmasses[i] = (g2->scalefactor)*(g2->m_masses[i]);
		for (int i = m2 ; i < nummasses ; i++)
			newmasses[i] = scalefactor*m_masses[i];
#endif // EXPCROSS
	}
	else
	{
#ifndef EXPCROSS
		for (int i = 0 ; i < nummasses ; i++)
		{
			float x = (float)uniDist.pickNumber();

			if (x < 0.5f)
				newmasses[i] = scalefactor*m_masses[i];
			else
				newmasses[i] = (g2->scalefactor)*(g2->m_masses[i]);
			if (newmasses[i] < 0)
				newmasses[i] = 0;
		}
#else
		UniformDistribution massPick(0, nummasses, factory->getRandomNumberGenerator());

		int m1 = (int)massPick.pickNumber();
		int m2 = (int)massPick.pickNumber();

		if (m1 > m2)
		{
			int tmp = m2;
			m2 = m1;
			m1 = tmp;
		}

		for (int i = 0 ; i < m1 ; i++)
			newmasses[i] = scalefactor*m_masses[i];
		for (int i = m1 ; i < m2 ; i++)
			newmasses[i] = (g2->scalefactor)*(g2->m_masses[i]);
		for (int i = m2 ; i < nummasses ; i++)
			newmasses[i] = scalefactor*m_masses[i];

		for (int i = 0 ; i < nummasses ; i++)
			if (newmasses[i] < 0)
				newmasses[i] = 0;
#endif // EXPCROSS
	}

	float newSheetValue = -1;

	if (m_sheetValue >= 0)
	{
		float x = (float)uniDist.pickNumber();

		if (x < 0.5f)
			newSheetValue = m_sheetValue;
		else
			newSheetValue = g2->m_sheetValue;
	}

#ifdef AUTODISTFRAC
	std::vector<float> newFractions(m_distanceFractions.size());

	for (int i = 0 ; i < m_distanceFractions.size() ; i++)
	{
		float x = (float)uniDist.pickNumber();

		if (x < 0.5f)
			newFractions[i] = m_distanceFractions[i];
		else
			newFractions[i] = g2->m_distanceFractions[i];
	}
	return new GridLensInversionGenomeBase(factory, newmasses, newFractions, newSheetValue);
#else
	return new GridLensInversionGenomeBase(factory, newmasses, newSheetValue);
#endif // AUTODISTFRAC
}

mogal::Genome *GridLensInversionGenomeBase::clone() const
{
#ifdef AUTODISTFRAC
	GridLensInversionGenomeBase *g = new GridLensInversionGenomeBase(factory, m_masses, m_distanceFractions, m_sheetValue);
#else
	GridLensInversionGenomeBase *g = new GridLensInversionGenomeBase(factory, m_masses, m_sheetValue);
#endif // AUTODISTFRAC
	g->setFitnessValues(fitnessvalues);
	g->setScaleFactor(scalefactor);
	g->setSheetFactor(m_sheetFactor);

	return g;
}

void GridLensInversionGenomeBase::mutate()
{
	int nummasses = m_masses.size();

	if (factory->allowNegativeValues())
	{
		float chanceMultiplier = factory->getChanceMultiplier();
		float chance = chanceMultiplier/((float)nummasses);
		SimpleUniformDistribution uniDist(factory->getRandomNumberGenerator());
		
		if (factory->useAbsoluteMutation())
		{
			for (int i = 0 ; i < nummasses ; i++)
			{
				if ((float)uniDist.pickNumber() < chance)
					m_masses[i] = (float)uniDist.pickNumber()*2.0f-1.0f;
			}
		}
		else
		{
			float mutationAmplitude = factory->getMutationAmplitude();

			for (int i = 0 ; i < nummasses ; i++)
			{
				if ((float)uniDist.pickNumber() < chance)
				{
					// allow larger mutations with smaller probablility
					// p(x) = (2/Pi)*1/(x^2+1)
					// cfr anomalous diffusion
					float p = (float)uniDist.pickNumber()*2.0f-1.0f;
					float x = TAN(p*(float)(CONST_PI/4.0))*mutationAmplitude;
					float y = x+m_masses[i];

					if (y < -1.0f)
						y = -1.0f;
					else if (y > 1.0f)
						y = 1.0f;
					
					m_masses[i] = y;
				}
			}
		}
	}
	else
	{
		float chanceMultiplier = factory->getChanceMultiplier();
		float chance = chanceMultiplier/((float)nummasses);
		SimpleUniformDistribution uniDist(factory->getRandomNumberGenerator());
		
		if (factory->useAbsoluteMutation())
		{
			for (int i = 0 ; i < nummasses ; i++)
			{
				if ((float)uniDist.pickNumber() < chance)
					m_masses[i] = (float)uniDist.pickNumber();
			}
		}
		else
		{
			float mutationAmplitude = factory->getMutationAmplitude();

			for (int i = 0 ; i < nummasses ; i++)
			{
				if ((float)uniDist.pickNumber() < chance)
				{
					// allow larger mutations with smaller probablility
					// p(x) = (2/Pi)*1/(x^2+1)
					// cfr anomalous diffusion
					float p = (float)uniDist.pickNumber()*2.0f-1.0f;
					float x = TAN(p*(float)(CONST_PI/4.0))*mutationAmplitude;
					float y = x+m_masses[i];

					if (y < 0)
						y = 0;
					else if (y > 1.0)
						y = 1.0;
					
					m_masses[i] = y;
				}
			}
		}
	}

	if (m_sheetValue >= 0)
	{
		float chanceMultiplier = factory->getChanceMultiplier();
		float chance = chanceMultiplier/((float)nummasses);
		SimpleUniformDistribution uniDist(factory->getRandomNumberGenerator());
		
		if (factory->useAbsoluteMutation())
		{
			if ((float)uniDist.pickNumber() < chance)
				m_sheetValue = (float)uniDist.pickNumber();
		}
		else
		{
			float mutationAmplitude = factory->getMutationAmplitude();

			if ((float)uniDist.pickNumber() < chance)
			{
				// allow larger mutations with smaller probablility
				// p(x) = (2/Pi)*1/(x^2+1)
				// cfr anomalous diffusion
				float p = (float)uniDist.pickNumber();
				float x = TAN(p*(float)(CONST_PI/4.0))*mutationAmplitude;
				float y = x+m_sheetValue;

				if (y < 0)
					y = 0;
				else if (y > 1.0f)
					y = 1.0f;
				
				m_sheetValue = y;
			}
		}
	}

	rescale();

#ifdef AUTODISTFRAC
	if (m_distanceFractions.size() > 0)
	{
		float chanceMultiplier = factory->getChanceMultiplier();
		//float chance = chanceMultiplier/((float)(m_distanceFractions.size()+1));
		float chance = chanceMultiplier/((float)(nummasses));
		SimpleUniformDistribution uniDist(factory->getRandomNumberGenerator());

		if (factory->useAbsoluteMutation())
		{
			for (int i = 0 ; i < m_distanceFractions.size() ; i++)
			{
				if ((float)uniDist.pickNumber() < chance)
					m_distanceFractions[i] = (float)uniDist.pickNumber();
			}
		}
		else
		{
			float mutationAmplitude = factory->getMutationAmplitude();

			for (int i = 0 ; i < m_distanceFractions.size() ; i++)
			{
				if ((float)uniDist.pickNumber() < chance)
				{
					// allow larger mutations with smaller probablility
					// p(x) = (2/Pi)*1/(x^2+1)
					// cfr anomalous diffusion
					float p = (float)uniDist.pickNumber()*2.0f-1.0f;
					float x = TAN(p*(float)(CONST_PI/4.0))*mutationAmplitude;
					float y = x+m_distanceFractions[i];

					if (y <= 0)
						y = uniDist.pickNumber();
					else if (y >= 1.0)
						y = uniDist.pickNumber();
					
					m_distanceFractions[i] = y;
				}
			}
		}
	}
#endif // AUTODISTFRAC
}

std::string GridLensInversionGenomeBase::getFitnessDescription() const
{
	char str[1024];
	char *p = str;
	int num = getFactory()->getNumberOfFitnessComponents();

	for (int i = 0 ; i < num ; i++)
	{
		sprintf(p,"%g ",(double)fitnessvalues[i]);
		p += strlen(p);
	}
	return std::string(str);
}
	
void GridLensInversionGenomeBase::rescale()
{
	int nummasses = m_masses.size();

	if (factory->allowNegativeValues())
	{
		float maxval = 0;
		
		for (int i = 0 ; i < nummasses ; i++)
		{
			if (ABS(maxval) < ABS(m_masses[i]))
				maxval = ABS(m_masses[i]);
		}

		if (maxval > 0)
		{
			float f = 0.5f/maxval;
			for (int i = 0 ; i < nummasses ; i++)
				m_masses[i] *= f;
		}
	}
	else
	{
		float maxval = 0;
		for (int i = 0 ; i < nummasses ; i++)
			maxval = MAX(maxval, m_masses[i]);
		
		if (maxval > 0)
		{
			float f = 0.5f/maxval;
			for (int i = 0 ; i < nummasses ; i++)
			{
				m_masses[i] *= f;
				if (m_masses[i] < 0)
					m_masses[i] = 0;
			}
		}
	}
}

// TODO: this should use the basis functions stored in gafactoryparams
GravitationalLens *GridLensInversionGenomeBase::createLens(double *totalmass, std::string &errstr) const
{
#if 0
	int nummasses = m_masses.size();
	const GridLensInversionGAFactoryParams *params = (const GridLensInversionGAFactoryParams *)factory->getCurrentParameters();
	const std::list<GridSquare> &squares = params->getGridSquares();
	std::list<GridSquare>::const_iterator it = squares.begin();
	const float *weights = factory->getMassWeights();
	double massscale = params->getMassScale();
	double tm = 0;
	GravitationalLens *lens = 0;
	
	if (params->getBasisFunctionType() == GridLensInversionGAFactoryParams::PlummerBasis)
	{
		std::list<PlummerLensInfo> newplummers;
	
		for (int i = 0 ; i < nummasses ; i++, it++)
		{
			double m = massscale*(double)m_masses[i]*(double)weights[i]*(double)scalefactor;
			newplummers.push_back(PlummerLensInfo(m, (*it).getSize(), (*it).getCenter()));
			tm += m;
		}
	
		MultiplePlummerLensParams lensparams(newplummers);
		lens = new MultiplePlummerLens();

		if (!lens->init(params->getD_d(), &lensparams))
		{
			errstr = lens->getErrorString();
			delete lens;
			return 0;
		}
	}
	else if (params->getBasisFunctionType() == GridLensInversionGAFactoryParams::GaussBasis)
	{
		std::list<GaussLensInfo> newgaussians;
	
		for (int i = 0 ; i < nummasses ; i++, it++)
		{
			double m = massscale*(double)m_masses[i]*(double)weights[i]*(double)scalefactor;
			newgaussians.push_back(GaussLensInfo(m, (*it).getSize(), (*it).getCenter()));
			tm += m;
		}
	
		MultipleGaussLensParams lensparams(newgaussians);
		lens = new MultipleGaussLens();

		if (!lens->init(params->getD_d(), &lensparams))
		{
			errstr = lens->getErrorString();
			delete lens;
			return 0;
		}
	}
	else // Multiple squares
	{
		std::list<SquareLensInfo> newsquares;
	
		for (int i = 0 ; i < nummasses ; i++, it++)
		{
			double m = massscale*(double)m_masses[i]*(double)weights[i]*(double)scalefactor;
			newsquares.push_back(SquareLensInfo(m, (*it).getSize(), (*it).getCenter()));
			tm += m;
		}
	
		MultipleSquareLensParams lensparams(newsquares);
		lens = new MultipleSquareLens();

		if (!lens->init(params->getD_d(), &lensparams))
		{
			errstr = lens->getErrorString();
			delete lens;
			return 0;
		}
	}
	*totalmass = tm;

	return lens;
#else
#ifdef AUTODISTFRAC
	const std::vector<float> &fixedDistFrac = factory->getDistanceFractions();
	int groupSize = factory->getImagesGroupSize();
	int count = 0;
	
	for (int i = 0 ; i < fixedDistFrac.size() ; i += groupSize)
	{
		std::cout << "Distance fraction for source (group) " << (i+1) << ": ";
		if (fixedDistFrac[i] != 1)
			std::cout << "" << fixedDistFrac[i] << "(fixed)" << std::endl;
		else
		{
			std::cout << "" << ((m_distanceFractions[count]*(1.0f-AUTODISTFRAC_MINVALUE))+AUTODISTFRAC_MINVALUE) << "(auto)" << std::endl;
			count++;
		}
	}

#endif // AUTODISTFRAC
	if (!factory->useLoopSheet())
		return factory->createLens(m_masses, m_sheetValue, scalefactor, totalmass, errstr);
	return factory->createLens(m_masses, m_sheetFactor/factory->getSheetScale(), scalefactor, totalmass, errstr);
#endif
}

void GridLensInversionGenomeBase::initializeNewCalculation()
{
	GridLensInversionGAFactoryBase *pFactory = getFactory();
	DeflectionMatrix *pDeflectionMatrix = pFactory->getDeflectionMatrix();
	BackProjectMatrixNew *pTotalBackProjectMatrix = pFactory->getTotalBackProjectMatrix();
	BackProjectMatrixNew *pShortBackProjectMatrix = pFactory->getShortBackProjectMatrix();

#ifdef AUTODISTFRAC
	const std::vector<float> &fixedDistFrac = pFactory->getDistanceFractions();
	int groupSize = pFactory->getImagesGroupSize();
	std::vector<float> fractions(fixedDistFrac.size());
	int count = 0;
	
	for (int i = 0 ; i < fixedDistFrac.size() ; i += groupSize)
	{
		if (fixedDistFrac[i] != 1)
		{
			for (int j = 0 ; j < groupSize ; j++)
				fractions[i+j] = fixedDistFrac[i+j]; // TODO: force this to the fraction of 'i' ??
		}
		else
		{
			for (int j = 0 ; j < groupSize ; j++)
				fractions[i+j] = (m_distanceFractions[count]*(1.0f-AUTODISTFRAC_MINVALUE)) + AUTODISTFRAC_MINVALUE;
			count++;
		}
	}

	pTotalBackProjectMatrix->setDistanceFractions(fractions);
	pShortBackProjectMatrix->setDistanceFractions(fractions);
#endif // AUTODISTFRAC

	pDeflectionMatrix->calculateBasisMatrixProducts(getMasses(), true, true, true);
	pTotalBackProjectMatrix->storeDeflectionMatrixResults();
	pShortBackProjectMatrix->storeDeflectionMatrixResults();
}

float GridLensInversionGenomeBase::calculateMassScaleFitness(float scaleFactor, float sheetScale)
{
	GridLensInversionGAFactoryBase *pFactory = getFactory();
	BackProjectMatrixNew *pShortBackProjectMatrix = pFactory->getShortBackProjectMatrix();
	LensFitnessObject &fitnessFunction = *(pFactory->getFitnessObject());

	pShortBackProjectMatrix->calculate(scaleFactor, sheetScale);
	if (fitnessFunction.shortNeedInverseMagnifications())
		pShortBackProjectMatrix->calculateInverseMagnifications(*(fitnessFunction.getShortInverseMagnificationFlags()));
	if (fitnessFunction.shortNeedShearComponents())
		pShortBackProjectMatrix->calculateShearComponents(*(fitnessFunction.getShortShearComponentFlags()));
	if (fitnessFunction.shortNeedConvergence())
		pShortBackProjectMatrix->calculateConvergence(*(fitnessFunction.getShortConvergenceFlags()));

	return fitnessFunction.calculateMassScaleFitness(*pShortBackProjectMatrix);
}

void GridLensInversionGenomeBase::calculateTotalFitness(float scaleFactor, float sheetScale, float *pFitnessValues)
{
	GridLensInversionGAFactoryBase *pFactory = getFactory();
	BackProjectMatrixNew *pTotalBackProjectMatrix = pFactory->getTotalBackProjectMatrix();
	LensFitnessObject &fitnessFunction = *(pFactory->getFitnessObject());
	
	pTotalBackProjectMatrix->calculate(scaleFactor, sheetScale);
	if (fitnessFunction.totalNeedInverseMagnifications())
		pTotalBackProjectMatrix->calculateInverseMagnifications(*(fitnessFunction.getTotalInverseMagnificationFlags()));
	if (fitnessFunction.totalNeedShearComponents())
		pTotalBackProjectMatrix->calculateShearComponents(*(fitnessFunction.getTotalShearComponentFlags()));
	if (fitnessFunction.totalNeedConvergence())
		pTotalBackProjectMatrix->calculateConvergence(*(fitnessFunction.getTotalConvergenceFlags()));

	fitnessFunction.calculateOverallFitness(*pTotalBackProjectMatrix, pFitnessValues);
}


} // end namespace


