/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#ifndef GRALE_GRIDTREE_H

#define GRALE_GRIDTREE_H

#include "graleconfig.h"
#include "realtype.h"
#include <vector>

// TODO: for debugging
#include <stdio.h>

namespace grale
{

class GRALE_IMPORTEXPORT GridNodeBase
{
public:
	enum GridNodeType { TreeNode, WeightNode };
protected:
	GridNodeBase(GridNodeType t, GridNodeBase *pParent, int childNumber, int level)	
	{ 
		m_nodeType = t; 
		m_pParent = pParent; 
		m_childNumber = childNumber; 
		m_level = level; 
	}
public:
	virtual ~GridNodeBase()							{ }
	virtual GridNodeBase *createCopy() const = 0;
	GridNodeType getNodeType() const					{ return m_nodeType; }
	GridNodeBase *getParent() const						{ return m_pParent; }
	int getChildNumber() const						{ return m_childNumber; }
	int getLevel() const							{ return m_level; }
	void setChildInfo(GridNodeBase *pParent, int childNumber)		{ m_pParent = pParent; m_childNumber = childNumber; }
	virtual void setLevel(int level)					{ m_level = level; }

	// TODO: for debugging
	virtual void print(int tab = 0) = 0;
private:
	GridNodeType m_nodeType;
	GridNodeBase *m_pParent;
	int m_childNumber;
	int m_level;
};

class GRALE_IMPORTEXPORT GridWeightNode : public GridNodeBase
{
public:
	GridWeightNode(float w, GridNodeBase *pParent, int childNumber, int level) : GridNodeBase(WeightNode, pParent, childNumber, level)
										{ m_weight = w; }
	~GridWeightNode()							{ }
	float getWeight() const							{ return m_weight; }
	void setWeight(float w)							{ m_weight = w; }
	GridNodeBase *createCopy() const					{ return new GridWeightNode(m_weight, 0, -1, 0); } // don't copy parent info to avoid double deletions

	// TODO: for debugging
	void print(int tab = 0)							{ for (int i = 0 ; i < tab ; i++) printf("   "); printf("%g (%d)\n",(double)m_weight, getLevel()); }
private:
	float m_weight;
};

class GRALE_IMPORTEXPORT GridTreeNode : public GridNodeBase
{
public:
	GridTreeNode(GridNodeBase *pParent, int childNumber, int level, int numChildren = 4) : GridNodeBase(TreeNode, pParent, childNumber, level)
	{ 
		m_pChild.resize(numChildren);
		for (int i = 0 ; i < numChildren ; i++) 
			m_pChild[i] = 0; 
	}
	
	~GridTreeNode()								
	{ 
		for (int i = 0 ; i < m_pChild.size() ; i++) 
			if (m_pChild[i]) 
				delete m_pChild[i]; 
	}

	int getNumberOfChildren() const									{ return m_pChild.size(); }

	bool isValid() const
	{ 
		for (int i = 0 ; i < m_pChild.size() ; i++) 
		{ 
			if (m_pChild[i] == 0) 
				return false; 
			if (m_pChild[i]->getNodeType() == TreeNode) 
			{ 
				const GridTreeNode *pNode = (const GridTreeNode *)m_pChild[i]; 
				
				if (!pNode->isValid()) 
					return false; 
			} 
		} 
		return true; 
	}

	void setNode(int i, GridNodeBase *pNode)
	{ 
		if (m_pChild[i]) 
			delete m_pChild[i]; 
		m_pChild[i] = pNode; 
		
		if (pNode) 
		{ 
			m_pChild[i]->setChildInfo(this, i); 
			m_pChild[i]->setLevel(getLevel()+1); 
		} 
	}

	void setNodeNoLevel(int i, GridNodeBase *pNode)	
	{ 
		if (m_pChild[i]) 
			delete m_pChild[i]; 
		m_pChild[i] = pNode; 
		
		if (pNode) 
		{ 
			m_pChild[i]->setChildInfo(this, i); 
		} 
	}

	GridNodeBase *createCopy() const					
	{ 
		GridTreeNode *pNewNode = new GridTreeNode(0, -1, 0, m_pChild.size()); 
		for (int i = 0 ; i < m_pChild.size() ; i++) 
		{ 
			GridNodeBase *pNode = m_pChild[i]; 
			if (pNode) 
				pNode = pNode->createCopy(); 
			pNewNode->setNode(i, pNode); 
		} 
		return pNewNode;
	}

	const GridNodeBase *getChild(int i) const				{ return m_pChild[i]; }
	GridNodeBase *getChild(int i)						{ return m_pChild[i]; }

	void setLevel(int level)
	{ 
		GridNodeBase::setLevel(level); 
		for (int i = 0 ; i < m_pChild.size() ; i++) 
			if (m_pChild[i]) 
				m_pChild[i]->setLevel(level+1); 
	}

	GridNodeBase *extractChild(int i)
	{ 
		GridNodeBase *pChild = m_pChild[i]; 
		m_pChild[i] = 0; 
		return pChild; 
	}
	
	// TODO: for debugging
	void print(int tab = 0)							{ for (int i = 0 ; i < tab ; i++) printf("   "); printf("Node (%d)\n", getLevel()); for (int i = 0 ; i < m_pChild.size() ; i++) { if (!m_pChild[i]) { for (int i = 0 ; i < tab+1 ; i++) printf("   "); printf("NULL\n"); } else { m_pChild[i]->print(tab+1); } } }
private:
	std::vector<GridNodeBase *> m_pChild;
};

} // end namespace

#endif // GRALE_GRIDTREE_H

