/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "fitsimage.h"
#include "wcscoordinates.h"
#include "rectangularcoordinates.h"
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	#include <fitsio.h>
#endif // GRALECONFIG_SUPPORT_CFITSIO
#ifndef WIN32
#include <unistd.h>
#endif // WIN32
#include <sys/stat.h>
#include <sys/types.h>
#include <stdint.h>
#include <string.h>

#include <iostream>

#include "debugnew.h"

namespace grale
{

FITSImage::FITSImage()
{
	m_pCoordinates = 0;
	m_pPixels = 0;
	m_width = 0;
	m_height = 0;
}

FITSImage::~FITSImage()
{
	if (m_pPixels)
		delete [] m_pPixels;
	if (m_pCoordinates)
		delete m_pCoordinates;
}

bool FITSImage::initNoCoordinates(int w, int h)
{
	if (m_pPixels)
	{
		setErrorString("An image is already loaded");
		return false;
	}

	m_width = w;
	m_height = h;
	m_pPixels = new double[m_width*m_height];
	
	memset(m_pPixels, 0, m_height*m_width*sizeof(double));

	m_header = std::string("");
	m_pCoordinates = 0;

	return true;
}

bool FITSImage::loadWCSCoordinates(const std::string &fname)
{
#ifndef GRALECONFIG_SUPPORT_CFITSIO
	setErrorString("No cfitsio support was compiled in");
	return false;
#else
	if (m_pPixels)
	{
		setErrorString("An image is already loaded");
		return false;
	}
	
	fitsfile *pFile;
	long naxes[2];
	int nfound;
	int status = 0;

	if (fits_open_file(&pFile, fname.c_str(), READONLY, &status))
	{
		std::cerr << "status = " << status << std::endl;
		setErrorString("Can't open FITS file");
		return false;
	}

	bool axesFound = false;
	
	while(!axesFound)
	{
		bool error = false;
		
		if (fits_read_keys_lng(pFile, "NAXIS", 1, 2, naxes, &nfound, &status))
			error = true;
		else if (nfound != 2)
			error = true;

		if (error) // try next HDU
		{
			int t;
			
			if (fits_movrel_hdu(pFile, 1, &t, &status))
			{
				fits_close_file(pFile, &status);
				setErrorString("Can't read image dimensions");
				return false;
			}
		}
		else
			axesFound = true;
	}

	m_width = naxes[0];
	m_height = naxes[1];
	m_pPixels = new double[m_width*m_height];
	
	double nullVal = 0;
	int anyNull = 0;
	
	if (fits_read_img(pFile, TDOUBLE, 1, (int)(naxes[0]*naxes[1]), &nullVal, m_pPixels, &anyNull, &status))
	{
		delete [] m_pPixels;
		m_pPixels = 0;
		fits_close_file(pFile, &status);
		setErrorString("Can't read FITS image data");
		return false;
	}

	char *pStr = 0;
	int nkeys = 0;
	if (fits_hdr2str(pFile, 0, 0, 0, &pStr, &nkeys, &status))
	{
		delete [] m_pPixels;
		m_pPixels = 0;
		fits_close_file(pFile, &status);
		setErrorString("Can't read FITS header string");
		return false;
	}
	
	std::string headerString(pStr);
	free(pStr);
	
	m_header = headerString;

	fits_close_file(pFile, &status);

	// Process the image itself

	WCSCoordinates *pCoordinates= new WCSCoordinates();

	if (!pCoordinates->init(headerString))
	{
		delete [] m_pPixels;
		m_pPixels = 0;
		setErrorString(std::string("Unable to create a WCS coordinate converter: ") + pCoordinates->getErrorString());
		delete pCoordinates;
		return false;
	}

	m_pCoordinates = pCoordinates;

	return true;
#endif // !GRALECONFIG_SUPPORT_CFITSIO
}

bool FITSImage::loadRectCoordinates(const std::string &fname)
{
#ifndef GRALECONFIG_SUPPORT_CFITSIO
	setErrorString("No cfitsio support was compiled in");
	return false;
#else
	if (m_pPixels)
	{
		setErrorString("An image is already loaded");
		return false;
	}
	
	fitsfile *pFile;
	long naxes[2];
	int nfound;
	int status = 0;

	if (fits_open_file(&pFile, fname.c_str(), READONLY, &status))
	{
		setErrorString("Can't open FITS file");
		return false;
	}

	if (fits_read_keys_lng(pFile, "NAXIS", 1, 2, naxes, &nfound, &status))
	{
		fits_close_file(pFile, &status);
		setErrorString("Can't read image dimensions");
		return false;
	}

	if (nfound != 2)
	{
		fits_close_file(pFile, &status);
		setErrorString("Read an invalid number of axes");
		return false;
	}

	m_width = naxes[0];
	m_height = naxes[1];
	m_pPixels = new double[m_width*m_height];
	
	double nullVal = 0;
	int anyNull = 0;
	
	if (fits_read_img(pFile, TDOUBLE, 1, (int)(naxes[0]*naxes[1]), &nullVal, m_pPixels, &anyNull, &status))
	{
		delete [] m_pPixels;
		m_pPixels = 0;
		fits_close_file(pFile, &status);
		setErrorString("Can't read FITS image data");
		return false;
	}

	char *pStr = 0;
	int nkeys = 0;
	if (fits_hdr2str(pFile, 0, 0, 0, &pStr, &nkeys, &status))
	{
		delete [] m_pPixels;
		m_pPixels = 0;
		fits_close_file(pFile, &status);
		setErrorString("Can't read FITS header string");
		return false;
	}
	
	std::string headerString(pStr);
	free(pStr);
	
	m_header = headerString;

	double crval[2];
	double cdelt[2];

	if (fits_read_keys_dbl(pFile, "CRVAL", 1, 2, crval, &nfound, &status))
	{
		fits_close_file(pFile, &status);
		setErrorString("Can't read simple coordinate system start");
		return false;
	}

	if (nfound != 2)
	{
		fits_close_file(pFile, &status);
		setErrorString("Can't read simple coordinate system start (wrong number of axes)");
		return false;
	}

	if (fits_read_keys_dbl(pFile, "CDELT", 1, 2, cdelt, &nfound, &status))
	{
		fits_close_file(pFile, &status);
		setErrorString("Can't read simple coordinate system increment");
		return false;
	}

	if (nfound != 2)
	{
		fits_close_file(pFile, &status);
		setErrorString("Can't read simple coordinate system increment (wrong number of axes)");
		return false;
	}

	fits_close_file(pFile, &status);

	Vector2D<double> bottomLeft(crval[0], crval[1]);
	Vector2D<double> topRight(crval[0] + (double)m_width*cdelt[0], crval[1] + (double)m_height*cdelt[1]);

	// TODO: unit?
	
	bottomLeft *= ANGLE_ARCSEC;
	topRight *= ANGLE_ARCSEC;

	m_pCoordinates = new RectangularCoordinates(bottomLeft, topRight, m_width, m_height);

	return true;
#endif // !GRALECONFIG_SUPPORT_CFITSIO
}

bool FITSImage::loadNoCoordinates(const std::string &fname)
{
#ifndef GRALECONFIG_SUPPORT_CFITSIO
	setErrorString("No cfitsio support was compiled in");
	return false;
#else
	if (m_pPixels)
	{
		setErrorString("An image is already loaded");
		return false;
	}
	
	fitsfile *pFile;
	long naxes[2];
	int nfound;
	int status = 0;

	if (fits_open_file(&pFile, fname.c_str(), READONLY, &status))
	{
		setErrorString("Can't open FITS file");
		return false;
	}

	bool axesFound = false;
	
	while(!axesFound)
	{
		bool error = false;
		
		if (fits_read_keys_lng(pFile, "NAXIS", 1, 2, naxes, &nfound, &status))
			error = true;
		else if (nfound != 2)
			error = true;

		if (error) // try next HDU
		{
			int t;
			
			if (fits_movrel_hdu(pFile, 1, &t, &status))
			{
				fits_close_file(pFile, &status);
				setErrorString("Can't read image dimensions");
				return false;
			}
		}
		else
			axesFound = true;
	}

	m_width = naxes[0];
	m_height = naxes[1];
	m_pPixels = new double[m_width*m_height];
	
	double nullVal = 0;
	int anyNull = 0;
	
	if (fits_read_img(pFile, TDOUBLE, 1, (int)(naxes[0]*naxes[1]), &nullVal, m_pPixels, &anyNull, &status))
	{
		delete [] m_pPixels;
		m_pPixels = 0;
		fits_close_file(pFile, &status);
		setErrorString("Can't read FITS image data");
		return false;
	}

	char *pStr = 0;
	int nkeys = 0;
	if (fits_hdr2str(pFile, 0, 0, 0, &pStr, &nkeys, &status))
	{
		delete [] m_pPixels;
		m_pPixels = 0;
		fits_close_file(pFile, &status);
		setErrorString("Can't read FITS header string");
		return false;
	}
	
	std::string headerString(pStr);
	free(pStr);
	
	m_header = headerString;

	fits_close_file(pFile, &status);

	m_pCoordinates = 0;

	return true;
#endif // !GRALECONFIG_SUPPORT_CFITSIO
}

FITSImage *FITSImage::createCopy() const
{
	FITSImage *pImg = new FITSImage();

	if (m_pPixels)
	{
		CoordinateInterface *pNewCoords = 0;

		if (m_pCoordinates)
		{
			pNewCoords = m_pCoordinates->createCopy();
			if (pNewCoords == 0)
			{
				setErrorString(m_pCoordinates->getErrorString());
				delete pImg;
				return 0;
			}
		}

		double *pPixels = new double[m_width*m_height];

		memcpy(pPixels, m_pPixels, m_width*m_height*sizeof(double));
		
		pImg->m_pCoordinates = pNewCoords;
		pImg->m_pPixels = pPixels;
		pImg->m_width = m_width;
		pImg->m_height = m_height;
		pImg->m_header = m_header;
	}

	return pImg;
}

bool FITSImage::save(const std::string &filename) const
{
#ifndef GRALECONFIG_SUPPORT_CFITSIO
	setErrorString("No cfitsio support was compiled in");
	return false;
#else

	struct stat fileinfo;
	fitsfile *fptr;
	int status = 0;

	if (stat(filename.c_str(),&fileinfo) == 0) // file exists, remove it
	{
		if (unlink(filename.c_str()) != 0)
		{
			setErrorString("Can't delete the existing file");
			return false;
		}
	}
			
	if (fits_create_file(&fptr,filename.c_str(),&status))
	{
		//fits_report_error(stderr,status);
		setErrorString("Can't create the FITS file");
		return false;
	}
	
	int bitpix = DOUBLE_IMG;
	long naxis = 2;
	long naxes[2] = { (long)m_width, (long)m_height };
	
	if (fits_create_img(fptr, bitpix, naxis, naxes, &status))
	{
		fits_close_file(fptr,&status);
		setErrorString("Can't create the FITS image entry");
		return false;
	}

	if (fits_write_img(fptr, TDOUBLE, 1, m_width*m_height, m_pPixels, &status))
	{
		fits_close_file(fptr,&status);
		setErrorString("Can't write the FITS image data");
		return false;
	}

	bool done = false;
	int offset = 0;
	
#define FITS_KEYLEN 80
	
	while (!done)
	{
		if (offset >= m_header.size())
			done = true;
		else
		{
			std::string record = m_header.substr(offset, offset + FITS_KEYLEN);
			std::string key = record.substr(0, 8);
			
			if (key == std::string("SIMPLE  ") || key == std::string("BITPIX  ") ||
			    key == std::string("NAXIS   ") || key == std::string("NAXIS1  ") ||
			    key == std::string("NAXIS2  ") || key == std::string("EXTEND  ") ||
			    key == std::string("END     "))
			{
				// Ignore these keywords, they're already set
			}
			else
			{
				if (fits_write_record(fptr, record.c_str(), &status))
				{
					fits_close_file(fptr,&status);
					setErrorString("Can't write the header");
					return false;
				}
			}
		}
	
		offset += FITS_KEYLEN;
	}
	
	if (fits_close_file(fptr,&status))
	{
		//fits_report_error(stderr,status);
		setErrorString("Can't close the fits file");
		return false;
	}

	return true;
#endif // !GRALECONFIG_SUPPORT_CFITSIO	
}

} // end namespace

