/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#ifndef GRALE_IMAGEPLANE_H

#define GRALE_IMAGEPLANE_H

#include "graleconfig.h"
#include "vector2d.h"
#include "gridfunction.h"
//#include "randomnumbergenerator.h"
//#include "simpleuniformdistribution.h"
#include "lensplane.h"
#include <enut/networklayeraddress.h>
#include <serut/serializationinterface.h>
#include <stdint.h>
#include <stdio.h>
#include <vector>
#include <list>
#include <string>

namespace grale
{

class GravitationalLens;
class LensPlane;
class SourcePlane;
class OpenCLKernel;

class GRALE_IMPORTEXPORT ImagePlane : public errut::ErrorBase
{
public:
	ImagePlane(const GravitationalLens *pLens, SourcePlane *sourceplane);
	virtual ~ImagePlane();
	bool init(Vector2D<double> bottomleft,Vector2D<double> topright,int xpoints,int ypoints, OpenCLKernel *pCLKernel = 0);
	bool init(Vector2D<double> bottomleft,Vector2D<double> topright,int xpoints,int ypoints,
	          const nut::NetworkLayerAddress &serverAddress, uint16_t serverPort);
	bool init(LensPlane *lensplane);

	// Assume that for rank 0 initMPI is called when for the other ranks
	// runMPIHelper is called
	bool initMPI(Vector2D<double> bottomleft,Vector2D<double> topright,int xpoints,int ypoints);
	static bool runMPIHelper(std::string &errStr);

	bool isInit() const									{ if (m_betas.size() == 0) return false; return true; }
	const GravitationalLens *getLens() const						{ return m_pLens; }
	SourcePlane *getSourcePlane()								{ return sourceplane; }
	
	int getNumXPoints() const								{ return numx; }
	int getNumYPoints() const								{ return numy; }
	int getNumXPixels() const								{ return numx-1; }
	int getNumYPixels() const								{ return numy-1; }
	Vector2D<double> getBottomLeft() const							{ return bottomleft; }
	Vector2D<double> getTopRight() const							{ return topright; }
	double getXStep() const									{ return xstep; }
	double getYStep() const									{ return ystep; }
	
	double getImageIntensityAccurate(int xPixel, int yPixel, int subSamples = 10) const;
	double getSourceIntensityAccurate(int xPixel, int yPixel, int subSamples = 10) const;
		
	double getInverseMagnification(int xpos,int ypos) const					{ double axx = m_alphaxx[xpos+ypos*numx]; double ayy = m_alphayy[xpos+ypos*numx]; double axy = m_alphaxy[xpos+ypos*numx]; return (1.0-axx)*(1.0-ayy)-axy*axy; }
	void getInverseMagnificationComponents(int xpos,int ypos, double &kappa, double &gamma1,  double &gamma2) const				{ double axx = m_alphaxx[xpos+ypos*numx]; double ayy = m_alphayy[xpos+ypos*numx]; double axy = m_alphaxy[xpos+ypos*numx]; kappa = 0.5*(axx+ayy); gamma1 = 0.5*(axx-ayy); gamma2 = axy; }
	
	const std::list<Vector2D<double> > &getCriticalLinePoints() const			{ return criticallines; }
	const std::list<Vector2D<double> > &getCausticPoints() const				{ return caustics; }
	const std::list<std::list<Vector2D<double> > *> &getCriticalLineSegments() const	{ return criticallines2; }
	const std::list<std::list<Vector2D<double> > *> &getCausticSegments() const		{ return caustics2; }
	
	Vector2D<double> getIndexCoordinate(int xpos, int ypos) const;
	Vector2D<double> getPixelCoordinate(int xPixel, int yPixel) const;

	Vector2D<double> traceIndexCoordinate(int xpos, int ypos) const;
	bool getIndices(Vector2D<double> v,int *xpos,int *ypos) const;

	bool traceBeta(Vector2D<double> beta, std::list<Vector2D<double> > &thetaPoints, int numIterations = 4, bool check = true) const;
	bool traceTheta(Vector2D<double> theta, Vector2D<double> *pBeta) const;
	bool traceThetaApprox(Vector2D<double> theta, Vector2D<double> *pBeta) const;
	double getInverseMagnification(Vector2D<double> theta) const;
	double getInverseMagnificationApprox(Vector2D<double> theta) const;

	bool write(serut::SerializationInterface &si,bool writesources = false) const;
	static bool read(serut::SerializationInterface &si,ImagePlane **ip,std::string &errstr);
	static bool load(const std::string &fname,ImagePlane **ip,std::string &errstr);
	bool save(const std::string &fname,bool writesources = false) const;
protected:
	virtual void setFeedbackStatus(const std::string &msg) 					{ }
	virtual void setFeedbackPercentage(int pct)						{ }
private:
	class InternalLensPlane : public LensPlane
	{
	public:
		InternalLensPlane(const GravitationalLens *pLens, ImagePlane *pImgPlane) : LensPlane(pLens)
		{
			m_pImgPlane = pImgPlane;
		}

		~InternalLensPlane()
		{
		}

		void setFeedbackStatus(const std::string &msg) 
		{
			m_pImgPlane->setFeedbackStatus(msg);
		}

		virtual void setFeedbackPercentage(int pct)
		{
			m_pImgPlane->setFeedbackPercentage(pct);
		}
	private:
		ImagePlane *m_pImgPlane;
	};

	ImagePlane() /* : m_uniDist(&m_rndGen) */							{ }
	static double getTime();

	void searchCritCaust(double *inversemagnifications);
	bool refinePosition(Vector2D<double> beta, Vector2D<double> startTheta, Vector2D<double> &theta, int numIterations) const;
	bool calculateBeta(Vector2D<double> theta, Vector2D<double> *pBeta) const;
	Vector2D<double> approximateBeta(int xIdx, int yIdx, Vector2D<double> theta) const;
	double getTriangleIntensity(Vector2D<double> p1, Vector2D<double> p2, Vector2D<double> p3,
			            double axx1, double axx2, double axx3,
				    double ayy1, double ayy2, double ayy3,
				    double axy1, double axy2, double axy3,
				    int numPoints, bool useMagnification) const;

	GravitationalLens *m_pLens;
	SourcePlane *sourceplane;
	
	std::vector<Vector2D<double> > m_betas;
	std::vector<double> m_alphaxx;
	std::vector<double> m_alphayy;
	std::vector<double> m_alphaxy;
	GridFunction *m_pAlphaxxFunction;
	GridFunction *m_pAlphayyFunction;
	GridFunction *m_pAlphaxyFunction;
	std::list<Vector2D<double> > criticallines;
	std::list<Vector2D<double> > caustics;
	std::list<std::list<Vector2D<double> > *> criticallines2;
	std::list<std::list<Vector2D<double> > *> caustics2;
	
	Vector2D<double> bottomleft;
	Vector2D<double> topright;
	double xstep,ystep;
	int numx,numy;

	//RandomNumberGenerator m_rndGen;
	//SimpleUniformDistribution m_uniDist;
};

inline Vector2D<double> ImagePlane::getIndexCoordinate(int xpos,int ypos) const
{ 
	return Vector2D<double>((((double)xpos))*xstep+bottomleft.getX(),(((double)ypos))*ystep+bottomleft.getY());
}

inline Vector2D<double> ImagePlane::getPixelCoordinate(int xPixel, int yPixel) const
{
	return Vector2D<double>((((double)xPixel))*xstep+bottomleft.getX(),(((double)yPixel))*ystep+bottomleft.getY())+Vector2D<double>(xstep/2.0, ystep/2.0);
}

inline Vector2D<double> ImagePlane::traceIndexCoordinate(int xpos,int ypos) const
{ 
	return m_betas[xpos+ypos*numx];
}

inline bool ImagePlane::getIndices(Vector2D<double> v,int *xpos,int *ypos) const
{
	int x,y;

	x = (int)(((v.getX()-bottomleft.getX())/xstep)+0.5);
	y = (int)(((v.getY()-bottomleft.getY())/ystep)+0.5);

	if (x < 0 || y < 0 || x >= numx || y >= numy)
		return false;
	*xpos = x;
	*ypos = y;
	return true;
}

inline double ImagePlane::getInverseMagnificationApprox(Vector2D<double> theta) const
{
	double axx = (*m_pAlphaxxFunction)(theta);
	double ayy = (*m_pAlphayyFunction)(theta);
	double axy = (*m_pAlphaxyFunction)(theta);
	double inverseMagnification = ABS((1.0-axx)*(1.0-ayy)-axy*axy);

	return inverseMagnification;
}

} // end namespace

#endif // GRALE_IMAGEPLANE_H

