/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "rectangularintegrator.h"

#include "debugnew.h"

namespace grale
{
	
RectangularIntegrator::RectangularIntegrator(Vector2D<double> bottomLeft, Vector2D<double> topRight, int numX, int numY)
{
	m_bottomLeft = bottomLeft;
	m_topRight = topRight;
	m_numX = numX;
	m_numY = numY;

	if (m_numX < 1)
		m_numX = 1;
	if (m_numY < 1)
		m_numY = 1;

	m_xStep = (m_topRight.getX() - m_bottomLeft.getX())/(double)m_numX;
	m_yStep = (m_topRight.getY() - m_bottomLeft.getY())/(double)m_numY;
}

RectangularIntegrator::~RectangularIntegrator()
{
}

double RectangularIntegrator::integrate(Real2DFunction &f, bool abs) const
{
	double total = 0;
	int numX = getXSteps();
	int numY = getYSteps();
	double xStep = getXStepSize();
	double yStep = getYStepSize();
	double area = xStep*yStep;
	
	int i;
	double x;

	for (i = 0, x = getBottomLeft().getX() ; i < numX ; i++, x += xStep)
	{
		int j;
		double y;

		for (j = 0, y = getBottomLeft().getY() ; j < numY ; j++, y += yStep)
		{
			double sum = 0;

			double v1 = f(Vector2D<double>(x, y));
			double v2 = f(Vector2D<double>(x + xStep, y));
			double v3 = f(Vector2D<double>(x, y + yStep));
			double v4 = f(Vector2D<double>(x + xStep, y + yStep));

			if (abs)
				sum = ABS(v1) + ABS(v2) + ABS(v3) + ABS(v4);
			else
				sum = v1 + v2 + v3 + v4;

			sum /= 4.0;
			
			total += sum;
		}
	}
	
	total *= area;
	
	return total;
}

} // end namespace
