/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "monopolegrid.h"
#include "polygon2d.h"
#include "imagesdata.h"
#include <iostream>

#include "debugnew.h"

namespace grale
{

MonopoleGrid::MonopoleGrid()
{
}

MonopoleGrid::~MonopoleGrid()
{
}

bool MonopoleGrid::init(const std::list<ImagesData *> images, const Real2DDerivableFunction &target,
	                Vector2D<double> center, double areaSize, int subDiv, double zeroPointFactor)
{
	clear();

	// Create convex hulls for the images

	std::vector<Polygon2D<double> > convexHulls;
	std::list<ImagesData *>::const_iterator it;

	for (it = images.begin() ; it != images.end() ; it++)
	{
		ImagesData &imgDat = *(*it);
		
		for (int j = 0 ; j < imgDat.getNumberOfImages() ; j++)
		{
			std::list<Vector2D<double> > points;

			for (int k = 0 ; k < imgDat.getNumberOfImagePoints(j) ; k++)
				points.push_back(imgDat.getImagePointPosition(j, k));
						
			Polygon2D<double> poly;

			poly.init(points, true);
			convexHulls.push_back(poly);
		}
	}

	// place the basis functions

	m_bottomLeft = center-Vector2D<double>(areaSize, areaSize)/2.0;
	m_stepSize = areaSize/(subDiv-1);

	for (int u = 0 ; u < subDiv ; u++)
	{
		for (int v = 0 ; v < subDiv ; v++)
		{
			Vector2D<double> center = m_bottomLeft + Vector2D<double>(u, v)*m_stepSize;
			double overallMinPos = 1e30;
			bool fixedMinPos = false;

			for (int i = 0 ; !fixedMinPos && i < convexHulls.size()  ; i++)
			{
				double minPos = 1e30;

				if (convexHulls[i].isInside(center))
				{
					overallMinPos = 0;
					fixedMinPos = true;
				}
				else
				{
					minPos = SQRT(convexHulls[i].getDistanceSquared(center));
					overallMinPos = MIN(overallMinPos, minPos);
				}
			}

			if (overallMinPos >= 4.0*m_stepSize) // give us some room
			{
				double length = overallMinPos*0.99; // just a precaution
				double zeroPoint = (m_stepSize*zeroPointFactor)/length;
				
				m_monoPoleFunctions.push_back(MonopoleBasisFunction(zeroPoint, length));
				m_monoPoleWeights.push_back(1.0);
				m_monoPoleCenters.push_back(center);
			}
		}
	}
	
	// Got all the possible basis functions, now build the pixel dependencies so that we don't
	// have to iterate over all the basis functions when calculating the total effect of the
	// basis functions.
	// In the same loop, we'll also store the discrete version of the target.
	
	m_pixelDependencies.resize(subDiv*subDiv);
	m_targetMap.resize(subDiv*subDiv);
	m_targetGradientMap.resize(subDiv*subDiv);
	
	int index = 0;
	int numBasisFunctions = m_monoPoleCenters.size();
	
	for (int y = 0 ; y < subDiv ; y++)
	{
		for (int x = 0 ; x < subDiv ; x++, index++)
		{
			Vector2D<double> pixelCenter = m_bottomLeft + Vector2D<double>(x, y)*m_stepSize;

			for (int b = 0 ; b < numBasisFunctions ; b++)
			{
				Vector2D<double> basisCenter = m_monoPoleCenters[b];
			
				Vector2D<double> diff = pixelCenter-basisCenter;

				if (diff.getLength() < m_monoPoleFunctions[b].getLengthScale())
				{
					// In this case, the basis function has an effect at the
					// pixel position
					
					m_pixelDependencies[index].push_back(b);
				}
			}
			
			m_targetMap[index] = target(pixelCenter);
			m_targetGradientMap[index] = target.getGradient(pixelCenter);
		}
	}

	m_monoPoleMap.resize(subDiv*subDiv);
	m_monoPoleGradientMap.resize(subDiv*subDiv);
		
	m_subDiv = subDiv;
	
	return true;
}

void MonopoleGrid::prepareFitnessCalculation()
{
	int index = 0;
	for (int y = 0 ; y < m_subDiv ; y++)
	{
		for (int x = 0 ; x < m_subDiv ; x++, index++)
		{
			const std::vector<int> &basisIndices = m_pixelDependencies[index];
			Vector2D<double> pixelCenter = m_bottomLeft + Vector2D<double>(x, y)*m_stepSize;

			double value = 0;
			Vector2D<double> gradient(0, 0);

			for (int i = 0 ; i < basisIndices.size() ; i++)
			{
				int b = basisIndices[i];

				Vector2D<double> center = m_monoPoleCenters[b];
				Vector2D<double> diff = pixelCenter-center;

				value += m_monoPoleFunctions[b](diff) * m_monoPoleWeights[b];
				gradient += m_monoPoleFunctions[b].getGradient(diff) * m_monoPoleWeights[b];
			}

			m_monoPoleMap[index] = value;
			m_monoPoleGradientMap[index] = gradient;

			//std::cout << pixelCenter.getX() << " " << pixelCenter.getY() << " " << m_monoPoleGradientMap[index].getLength() << std::endl;
		}
		//std::cout << std::endl;
	}
}

void MonopoleGrid::calculateFitness(double *pFlatness, double *pNegative, double scale)
{
	double flatness = 0;
	double negative = 0;

	int numPixels = m_targetMap.size();

	for (int i = 0 ; i < numPixels ; i++)
	{

		double targetValue = m_targetMap[i];
		double monopoleValue = m_monoPoleMap[i] * scale;
		double totalDensity = targetValue-monopoleValue;
		Vector2D<double> gradientDiff = m_targetGradientMap[i] - m_monoPoleGradientMap[i] * scale;

		flatness += gradientDiff.getLengthSquared();

#if 0
		if (totalDensity < 0) // try to avoid going more negative than the original map
		{
			if (targetValue < 0)
			{
				if (monopoleValue > 0)
					negative += monopoleValue;
			}
			else
				negative += -totalDensity;
		}
#else
		if (totalDensity < 0) // try to avoid going more negative than the original map
			negative += -totalDensity;
#endif

#if 0
		double targetValue = m_targetMap[i];
		double monopoleValue = m_monoPoleMap[i] * scale;
		double totalDensity = targetValue-monopoleValue;

		flatness += totalDensity*totalDensity;

		if (totalDensity < 0) // try to avoid going more negative than the original map
		{
			if (targetValue < 0)
			{
				if (monopoleValue > 0)
					negative += monopoleValue;
			}
			else
				negative += -totalDensity;
		}
#endif
	}

	*pFlatness = flatness;
	*pNegative = negative;
}

void MonopoleGrid::clear()
{
	m_monoPoleFunctions.clear();
	m_monoPoleCenters.clear();
	m_monoPoleWeights.clear();
	m_pixelDependencies.clear();

	m_targetMap.clear();
	m_targetGradientMap.clear();
	m_monoPoleMap.clear();
	m_monoPoleGradientMap.clear();

	m_subDiv = 0;
	m_stepSize = 0;
	m_bottomLeft = Vector2D<double>(0,0);
}

#define MONOPOLEGRID_ID 0x4452474d

bool MonopoleGrid::write(serut::SerializationInterface &si) const
{
	if (!si.writeInt32(MONOPOLEGRID_ID))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeInt32(m_subDiv))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeDouble(m_stepSize))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeDoubles(m_bottomLeft.getComponents(), 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeInt32(m_monoPoleFunctions.size()))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	for (int i = 0 ; i < m_monoPoleFunctions.size() ; i++)
	{
		double values[2];

		values[0] = m_monoPoleFunctions[i].getZeroPoint();
		values[1] = m_monoPoleFunctions[i].getLengthScale();

		if (!si.writeDoubles(values, 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		if (!si.writeDoubles(m_monoPoleCenters[i].getComponents(), 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}
	}

	for (int i = 0 ; i < m_pixelDependencies.size() ; i++)
	{
		int numDepend = m_pixelDependencies[i].size();

		if (!si.writeInt32(numDepend))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		for (int j = 0 ; j < numDepend ; j++)
		{
			if (!si.writeInt32(m_pixelDependencies[i][j]))
			{
				setErrorString(si.getErrorString());
				return false;
			}
		}
	}

	if (!si.writeDoubles(m_targetMap))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	for (int i = 0 ; i < m_targetGradientMap.size() ; i++)
	{
		if (!si.writeDoubles(m_targetGradientMap[i].getComponents(), 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}
	}

	return true;
}

bool MonopoleGrid::read(serut::SerializationInterface &si)
{
	int32_t id;

	if (!si.readInt32(&id))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (id != MONOPOLEGRID_ID)
	{
		setErrorString("Read an invalid ID");
		return false;
	}

	int32_t subDiv;

	if (!si.readInt32(&subDiv))
	{
		setErrorString("Read an invalid subdivision number");
		return false;
	}

	double stepSize;

	if (!si.readDouble(&stepSize))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	Vector2D<double> bottomLeft;

	if (!si.readDoubles(bottomLeft.getComponents(), 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t numBasisFunctions;

	if (!si.readInt32(&numBasisFunctions))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	std::vector<MonopoleBasisFunction> basisFunctions;
	std::vector<Vector2D<double> > centers;

	for (int32_t i = 0 ; i < numBasisFunctions ; i++)
	{
		double values[2];
		Vector2D<double> center;

		if (!si.readDoubles(values, 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		basisFunctions.push_back(MonopoleBasisFunction(values[0], values[1]));

		if (!si.readDoubles(center.getComponents(), 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		centers.push_back(center);
	}

	int numPixels = subDiv*subDiv;

	std::vector<std::vector<int> > dependencies(numPixels);

	for (int i = 0 ; i < numPixels ; i++)
	{
		int32_t numDepend;

		if (!si.readInt32(&numDepend))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		dependencies[i].resize(numDepend);

		for (int32_t j = 0 ; j < numDepend ; j++)
		{
			int32_t value;

			if (!si.readInt32(&value))
			{
				setErrorString(si.getErrorString());
				return false;
			}

			dependencies[i][j] = value;
		}
	}

	std::vector<double> targetMap(numPixels);

	if (!si.readDoubles(targetMap))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	std::vector<Vector2D<double> > gradientMap(numPixels);

	for (int i = 0 ; i < numPixels ; i++)
	{
		Vector2D<double> gradient;

		if (!si.readDoubles(gradient.getComponents(), 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		gradientMap[i] = gradient;
	}

	// Ok, we've got it all, now store it in this instance

	clear();

	m_subDiv = subDiv;
	m_stepSize = stepSize;
	m_bottomLeft = bottomLeft;

	m_monoPoleFunctions = basisFunctions;
	m_monoPoleCenters = centers;

	m_monoPoleWeights.resize(m_monoPoleFunctions.size());
	for (int i = 0 ; i < m_monoPoleWeights.size() ; i++)
		m_monoPoleWeights[i] = 1;

	m_pixelDependencies = dependencies;
	m_targetMap = targetMap;
	m_targetGradientMap = gradientMap;

	m_monoPoleMap.resize(m_targetMap.size());
	m_monoPoleGradientMap.resize(m_targetMap.size());

	return true;
}

} // end namespace

