/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"

#ifdef GRALECONFIG_SUPPORT_CFITSIO

#include "imageplanefitswriter.h"
#include "imageplane.h"
#include <fitsio.h>
#ifndef WIN32
#include <unistd.h>
#endif // WIN32
#include <sys/stat.h>
#include <sys/types.h>
#include <stdint.h>

#include "debugnew.h"

namespace grale
{

bool ImagePlaneFITSWriter::write(ImagePlane *p, int subSamples)
{
	struct stat fileinfo;
	fitsfile *fptr;
	int status = 0;

	if (stat(filename.c_str(),&fileinfo) == 0) // file exists, remove it
	{
		if (unlink(filename.c_str()) != 0)
		{
			setErrorString("Can't delete the existing file");
			return false;
		}
	}
			
	if (fits_create_file(&fptr,filename.c_str(),&status))
	{
		//fits_report_error(stderr,status);
		setErrorString("Can't create the FITS file");
		return false;
	}
	
	int numx,numy;
	
	numx = p->getNumXPixels();
	numy = p->getNumYPixels();

	int bitpix = DOUBLE_IMG;
	long naxis = 2;
	long naxes[2] = {numx,numy};
	
	if (fits_create_img(fptr,bitpix,naxis,naxes,&status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't create the FITS image entry");
		return false;
	}

	int datalen = numx*numy;
	double *data = new double[datalen];
	long *data2 = 0;
	int index = 0;
	int x,y;

	if (m_plotImages)
	{
		for (y = 0 ; y < numy ; y++)
		{
			for (x = 0 ; x < numx ; x++)
			{
				double value = p->getImageIntensityAccurate(x, y, subSamples);
				
				data[index] = value;
				index++;
			}
		}
	}
	else // sources
	{
		for (y = 0 ; y < numy ; y++)
		{
			for (x = 0 ; x < numx ; x++)
			{
				double value = p->getSourceIntensityAccurate(x, y, subSamples);
				
				data[index] = value;
				index++;
			}
		}
	}

	if (fits_write_img(fptr,TDOUBLE,1,numx*numy,data,&status))
	{
		delete [] data;
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write the FITS image data");
		return false;
	}

	delete [] data;

	double stepx = p->getXStep()/scale;
	double stepy = p->getYStep()/scale;
	double startx = p->getBottomLeft().getX()/scale;
	double starty= p->getBottomLeft().getY()/scale;
	
	if (fits_update_key(fptr, TDOUBLE, "CRVAL1", &startx," ", &status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write the FITS CRVAL1 key");
		return false;
	}
	if (fits_update_key(fptr, TDOUBLE, "CRVAL2", &starty," ", &status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write the FITS CRVAL2 key");
		return false;
	}	
	if (fits_update_key(fptr, TDOUBLE, "CDELT1", &stepx," ", &status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write the FITS CDELT1 key");
		return false;
	}
	if (fits_update_key(fptr, TDOUBLE, "CDELT2", &stepy," ", &status))
	{
		//fits_report_error(stderr,status);
		fits_close_file(fptr,&status);
		setErrorString("Can't write the FITS CDELT2 key");
		return false;
	}
	
	if (fits_close_file(fptr,&status))
	{
		//fits_report_error(stderr,status);
		setErrorString("Can't close the fits file");
		return false;
	}

	return true;
}

} // end namespace

#endif // GRALECONFIG_SUPPORT_CFITSIO

