/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "real2dfunctiongnuplotplotter.h"
//#include <unistd.h>
#include <sys/types.h>
#include <stdint.h>
#include <sys/stat.h>
#include <stdio.h>

#include "debugnew.h"

namespace grale
{

Real2DFunctionGnuPlotPlotter::Real2DFunctionGnuPlotPlotter(const std::string &fname,double xs,double ys,double zs)
{
	filename = fname;
	xscale = xs;
	yscale = ys;
	zscale = zs;
	color = true;
	flag3d = true;
	axes = true;
	zsize = 0;
}

Real2DFunctionGnuPlotPlotter::~Real2DFunctionGnuPlotPlotter()
{
}

bool Real2DFunctionGnuPlotPlotter::plot(Real2DFunction &g,Vector2D<double> bottomleft,Vector2D<double> topright,int numxsteps,int numysteps)
{
	if (numxsteps < 2 || numysteps < 2)
	{
		setErrorString("The number of X and Y steps should be at least 2");
		return false;
	}
	
	FILE *f;
	
	f = fopen(filename.c_str(),"w");
	if (f == 0)
	{
		setErrorString("Can't open file for writing");
		return false;
	}
	
	double pointsize = 75.0/(double)(numxsteps);
		
	if (!color)
		fprintf(f,"set terminal postscript enhanced monochrome solid\n");
	else
		fprintf(f,"set terminal postscript enhanced color solid\n");
	fprintf(f,"set size 1.0,1.3\n");
	if (!flag3d)
	{
		fprintf(f,"set pointsize %g\n",pointsize);
		fprintf(f,"set view map\n");
	}
	else
	{
		fprintf(f,"set hidden3d\n");
		fprintf(f,"set view 45,45\n");
		fprintf(f,"set ticslevel 0\n");
	}
	if (!axes)
	{
		fprintf(f,"unset xtics\n");
		fprintf(f,"unset ytics\n");
		fprintf(f,"unset ztics\n");
	}
	fprintf(f,"set noborder\n");
	if (!color)
		fprintf(f,"set palette defined ( 0 'white',1 'black')\n");
	if (xlabel.length() > 0)
		fprintf(f,"set xlabel '%s'\n",xlabel.c_str());
	if (ylabel.length() > 0)
		fprintf(f,"set ylabel '%s'\n",ylabel.c_str());
	
	if (!flag3d)
		fprintf(f,"splot '-' notitle with points pt 47 palette\n");
	else
	{
		if (!color)
		{
			if (zsize == 0)
				fprintf(f,"splot '-' notitle with lines\n");
			else
				fprintf(f,"splot [][][0:%g] '-' notitle with lines\n",(double)zsize);
		}
		else
		{
			if (zsize == 0)
				fprintf(f,"splot '-' notitle with lines palette\n");
			else
				fprintf(f,"splot [][][0:%g] '-' notitle with lines palette\n",(double)zsize);
		}
	}
	int x,y;
	double startx = bottomleft.getX();
	double starty = bottomleft.getY();
	double xstep = (topright.getX()-bottomleft.getX())/((double)(numxsteps-1));
	double ystep = (topright.getY()-bottomleft.getY())/((double)(numysteps-1));

	if (!flag3d)
	{
		for (y = 0 ; y < numysteps ; y++)
		{
			for (x = 0 ; x < numxsteps ; x++)
			{
				double x2 = startx+xstep*(double)x;
				double y2 = starty+ystep*(double)y;
				double z = g(Vector2D<double>(x2,y2));
				x2 /= xscale;
				y2 /= yscale;
				z /= zscale;
				fprintf(f,"%.15Lg\t%.15Lg\t%.15Lg\n",(long double)x2,(long double)y2,(long double)z);
			}
		}
	}
	else
	{
		for (y = 0 ; y < numysteps-1 ; y++)
		{
			for (x = 0 ; x < numxsteps-1 ; x++)
			{
				double x2 = startx+xstep*(double)x;
				double y2 = starty+ystep*(double)y;
				double z1 = g(Vector2D<double>(x2,y2));
				double z2 = g(Vector2D<double>(x2+xstep,y2));
				double z3 = g(Vector2D<double>(x2+xstep,y2+ystep));
				double z4 = g(Vector2D<double>(x2,y2+xstep));
				x2 /= xscale;
				y2 /= yscale;
				z1 /= zscale;
				z2 /= zscale;
				z3 /= zscale;
				z4 /= zscale;
				fprintf(f,"%.15Lg\t%.15Lg\t%.15Lg\n",(long double)x2,(long double)y2,(long double)z1);
			}
			fprintf(f,"\n");
		}
	}
	
	fprintf(f,"e\n");
	fclose(f);
	
	return true;
}

} // end namespace

