/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "ellipticalsource.h"

#include "debugnew.h"

namespace grale
{

EllipticalSource::EllipticalSource(Vector2D<double> angularpos, double angular_axis, double excentricity, double rotangle, double brightnessScale) : SourceImage(SourceImage::Ellipse, angularpos, rotangle, brightnessScale)
{
	fade = false;
	A = angular_axis;
	e = excentricity;
	B = A*SQRT(1.0-e*e);
}

EllipticalSource::EllipticalSource() : SourceImage(SourceImage::Ellipse,Vector2D<double>(0,0), 0, 1.0)
{
	fade = false;
	A = 0;
	e = 0;
	B = 0;
}

EllipticalSource::~EllipticalSource()
{
}

SourceImage *EllipticalSource::createCopy() const
{
	EllipticalSource *src = new EllipticalSource(getAngularPosition(), A, e, getAngle(), getBrightnessScale());
	src->fade = fade;
	return src;
}

bool EllipticalSource::writeInternal(serut::SerializationInterface &si) const
{
	double values[2];

	values[0] = A;
	values[1] = e;

	if (!si.writeDoubles(values,2))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	
	int32_t fadeval = 0;
	if (fade)
		fadeval = 1;

	if (!si.writeInt32(fadeval))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	return true;
}

bool EllipticalSource::readInternal(serut::SerializationInterface &si)
{
	double values[2];
	int32_t fadeval;

	if (!si.readDoubles(values,2))
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (!si.readInt32(&fadeval))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	A = values[0];
	e = values[1];
	B = A*SQRT(1.0-e*e);
	if (fadeval == 0)
		fade = false;
	else
		fade = true;
	return true;
}

} // end namespace

